      Program Appendix_7_2_1
C
C
C  This is a main program for the example 5. This example, in parallel,
C  solves a system of linear equations in the form [A]{X}={B}
C  by the conjugate gradient method. This example uses a parallel
C  lock to handle all the critical sections, and uses two parallel
C  events to handle task synchronization.
C  
C  global variables      
C
      INTEGER*4 Order
      PARAMETER (Order=200)
      INTEGER*4 NumberOfTasks
      PARAMETER (NumberOfTasks=4)
      INTEGER*4 GrainSize
      PARAMETER (GrainSize=1000)
      INTEGER*4 N,LockID,EventA,EventB,Subset1,Subset23,Subset4
      INTEGER*4 Subset56,Subset78,Subset9,Outstanding,Tasks
      INTEGER*4 J4TEMP,K4TEMP,SEED,TaskID(NumberOfTasks)
      INTEGER*4 TaskIndex(NumberOfTasks)
      REAL*4 A(Order,Order),X(Order),B(Order),R(Order),P(Order)
      REAL*4 WORKold(NumberOfTasks),WORKnew(NumberOfTasks)
      REAL*4 WORKpap(NumberOfTasks),Alpha,Beta,Tolerance
      LOGICAL*4 Done
C
C  declare dispatched subroutine
C
      EXTERNAL SubA
      DATA Tolerance/0.1/,SEED/88888888/
C
C  generate data in a sequential procedure
C
      WRITE(*,'('' >>> Generate data ...'')')
      Tasks=NumberOfTasks
      N=Order
      DO J4TEMP=1,N*N
         A(J4TEMP,1)=RAN(SEED)
      END DO
      DO J4TEMP=1,N
         DO K4TEMP=1,N
            A(J4TEMP,J4TEMP)=A(J4TEMP,J4TEMP)+A(J4TEMP,K4TEMP)
         END DO
         A(J4TEMP,J4TEMP)=A(J4TEMP,J4TEMP)*5.0
      END DO
      DO J4TEMP=1,N
         B(J4TEMP)=0.0
         DO K4TEMP=1,N
            B(J4TEMP)=B(J4TEMP)+A(J4TEMP,K4TEMP)
         END DO
      END DO
C
C  create parallel lock
C
      CALL CreateParallelLock(LockID)
C
C  create parallel events
C
      CALL CreateParallelEvent(EventA)
      CALL CreateParallelEvent(EventB)
C
C  determine length of subsets
C
      Subset1=MIN0(GrainSize,Order/NumberOfTasks)
      IF(Subset1.EQ.0) Subset1=1
      Subset23=MIN0(GrainSize/4,Order/NumberOfTasks)
      IF(Subset23.EQ.0) Subset23=1
      Subset4=MIN0(GrainSize/3/Order,Order/NumberOfTasks)
      IF(Subset4.EQ.0) Subset4=1
      Subset56=MIN0(GrainSize/(3*Order+2),Order/NumberOfTasks)
      IF(Subset56.EQ.0) Subset56=1
      Subset78=MIN0(GrainSize/2,Order/NumberOfTasks)
      IF(Subset78.EQ.0) Subset78=1
      Subset9=MIN0(GrainSize/2,Order/NumberOfTasks)
      IF(Subset9.EQ.0) Subset9=1
C
C  initialize
C
      Outstanding=1
      CALL InitializeParallelEvent(EventA,Tasks)
C
C  solve in parallel
C
      WRITE(*,'('' >>> Solve in parallel ...'')')
      DO J4TEMP=1,NumberOfTasks-1
         TaskIndex(J4TEMP)=J4TEMP
         CALL Dispatch_25(TaskID(J4TEMP),SubA,A,X,B,N,R,P,
     &                    WORKold,WORKnew,WORKpap,Alpha,Beta,
     &                    Tasks,LockID,EventA,EventB,Subset1,Subset23,
     &                    Subset4,Subset56,Subset78,Subset9,Outstanding,
     &                    TaskIndex(J4TEMP),Tolerance,Done)
      END DO
      CALL SubA(A,X,B,N,R,P,WORKold,WORKnew,WORKpap,Alpha,Beta,
     &          Tasks,LockID,EventA,EventB,Subset1,Subset23,
     &          Subset4,Subset56,Subset78,Subset9,Outstanding,
     &          Tasks,Tolerance,Done)
C
C  terminate tasks
C
      CALL WaitForAndTerminateAllTasks
C
C  delete parallel lock
C
      CALL DeleteParallelLock(LockID)
C
C  delete parallel events
C
      CALL DeleteParallelEvent(EventA)
      CALL DeleteParallelEvent(EventB)
C
C  output
C
      WRITE(*,'(''      X(  1)= '',F7.4)') X(  1)
      WRITE(*,'(''      X( 50)= '',F7.4)') X( 50)
      WRITE(*,'(''      X(100)= '',F7.4)') X(100)
      WRITE(*,'(''      X(150)= '',F7.4)') X(150)
      WRITE(*,'(''      X(200)= '',F7.4)') X(200)
C
      STOP
      END
      
      
      SUBROUTINE SubA(A,X,B,N,R,P,WORKold,WORKnew,WORKpap,Alpha,Beta,
     &                Tasks,LockID,EventA,EventB,Subset1,Subset23,
     &                Subset4,Subset56,Subset78,Subset9,Outstanding,
     &                TaskIndex,Tolerance,Done)
C
C
C  routine to solve a system of linear equations in the form
C  [A]{X}={B} by the Conjugate gradient method, in which [A] is a
C  positive definite matrix of order (NxN), {B} is the right side
C  vector, and {X} is the solution to be determined.
C  (A)FORTRAN CALL: CALL SubA(A,X,B,N,R,P,WORKold,WORKnew,WORKpap,
C                             Alpha,Beta,Tasks,LockID,EventA,EventB,
C                             Subset1,Subset23,Subset4,
C                             Subset56,Subset78,Subset9,Outstanding,
C                             TaskIndex,Tolerance,Done)
C     1.A: <R4> left side matrix, dimension(N,N)
C     2.X: <R4> return solution, dimension(N)
C     3.B: <R4> right side vector, dimension(N)
C     4.N: <I4> order
C     5.R: <R4> working array for residual, dimension(N)
C     6.P: <R4> working array for directional vector, dimension(N)
C     7.WORKold: <R4> working array for the residual norm in
C                     previous iteration, dimension(Tasks)
C     8.WORKnew: <R4> working array for the residual norm in
C                     current iteration, dimension(Tasks)
C     9.WORKpap: <R4> working array for {P}[A]{P}, dimension(Tasks)
C    10.Alpha,Beta: <R4> working array for the coefficients alpha and beta
C    11.Tasks: <I4> number of tasks
C    12.LockID: <I4> lock identifier
C    13.EventA: <I4> identifier of the first event, which must be
C                    initialized in the caller's program
C    14.EventB: <I4> identifier of the second event
C    15.Subset1: <I4> length of subset for computing function 1
C    16.Subset23: <I4> length of subset for computing functions 2 and 3
C    17.Subset4: <I4> length of subset for computing function 4
C    18.Subset56: <I4> length of subset for computing functions 5 and 6
C    19.Subset78: <I4> length of subset for computing functions 7 and 8
C    20.Subset9: <I4> length of subset for computing function 9
C    21.Outstanding: <I4> outstanding element that must be initialized
C                         to one in the caller
C    22.TaskIndex: <I4> task ordinal number
C    23.Tolerance: <R4> tolerance to terminate computations
C    24.Done: <L4> working variable for termination of computation
C
C  dummy arguments
C
      INTEGER*4 N,Tasks,LockID,EventA,EventB,Subset1,Subset23,Subset4
      INTEGER*4 Subset56,Subset78,Subset9,Outstanding,TaskIndex
      REAL*4 A(N,N),X(N),B(N),R(N),P(N),WORKold(Tasks),WORKnew(Tasks)
      REAL*4 WORKpap(Tasks),Alpha,Beta,Tolerance
      LOGICAL*4 Done
C
C  private variables
C
      INTEGER*4 J4TEMP$,K4TEMP$,M4TEMP$
      LOGICAL*4 YES$
      AUTOMATIC J4TEMP$,K4TEMP$,M4TEMP$,YES$
C
C  set initial value
C
      Done=.False.
C
C  [critical section] --- function 1
C  distribute subsets
C
10    CALL LockonParallelLock(LockID)
      IF(Outstanding.GT.N) THEN
         CALL UnlockParallelLock(LockID)
         GO TO 20
      END IF
      J4TEMP$=Outstanding
      Outstanding=Outstanding+Subset1
      CALL UnlockParallelLock(LockID)
C
C  [parallelism] --- function 1
C  initialize solution vector, Size=Subset1
C
      DO K4TEMP$=J4TEMP$,MIN0(J4TEMP$+Subset1-1,N)
         X(K4TEMP$)=0.0
      END DO
      GO TO 10
C
C  [task synchronization] --- function 1
C  complete a cycle of parallel event A
C
20    CALL PostParallelEvent(EventA)
      CALL IfCompleteParallelEvent(EventA,Yes$)
C
C  [sequential fragment] --- function 1
C  initialize parallel event B and oustanding element
C
      IF(.NOT.YES$) THEN
         CALL InitializeParallelEvent(EventB,Tasks)
         Outstanding=1
         CALL CompleteParallelEvent(EventA)
      END IF
C
C  [critical section] --- functions 2 and 3
C  distribute subsets
C
      WORKold(TaskIndex)=0.0
30    CALL LockonParallelLock(LockID)
      IF(Outstanding.GT.N) THEN
         CALL UnlockParallelLock(LockID)
         GO TO 40
      END IF
      J4TEMP$=Outstanding
      Outstanding=Outstanding+Subset23
      CALL UnlockParallelLock(LockID)
C
C  [parallelism] --- functions 2 and 3
C  compute residual and norm, Size=4*Subset23
C
      DO K4TEMP$=J4TEMP$,MIN0(J4TEMP$+Subset23-1,N)
         R(K4TEMP$)=B(K4TEMP$)
         P(K4TEMP$)=R(K4TEMP$)
         WORKold(TaskIndex)=WORKold(TaskIndex)+R(K4TEMP$)*R(K4TEMP$)
      END DO
      GO TO 30
C
C  [task synchronization] --- functions 2 and 3
C  complete a cycle of event B
C
40    CALL PostParallelEvent(EventB)
      CALL IfCompleteParallelEvent(EventB,YES$)
C
C  [sequential fragment] --- functions 2 and 3
C  sum the residual norm, and initialize event A and outstanding element
C
      IF(.NOT.YES$) THEN
         DO J4TEMP$=2,Tasks
            WORKold(1)=WORKold(1)+WORKold(J4TEMP$)
         END DO
         CALL InitializeParallelEvent(EventA,Tasks)
         Outstanding=1
         CALL CompleteParallelEvent(EventB)
      END IF
C
C  [critical section] --- function 4
C  distribute subsets
C
      WORKpap(TaskIndex)=0.0
50    CALL LockonParallelLock(LockID)
      IF(Outstanding.GT.N) THEN
         CALL UnlockParallelLock(LockID)
         GO TO 60
      END IF
      J4TEMP$=Outstanding
      Outstanding=Outstanding+Subset4
      CALL UnlockParallelLock(LockID)
C
C  [parallelism] --- function 4
C  compute a partial sum of {P}[A]{P}, size=3*N*Subset4
C
      DO K4TEMP$=J4TEMP$,MIN0(J4TEMP$+Subset4-1,N)
         DO M4TEMP$=1,N
            WORKpap(TaskIndex)=WORKpap(TaskIndex)+P(K4TEMP$)*
     &                                            A(K4TEMP$,M4TEMP$)*
     &                                            P(M4TEMP$)
         END DO
      END DO
      GO TO 50
C
C  [task synchronization] --- function 4
C  complete a cycle of event A
C
60    CALL PostParallelEvent(EventA)
      CALL IfCompleteParallelEvent(EventA,Yes$)
C
C  [sequential fragment] --- function 4
C  sum {P}[A]{P}, initialize event B and outstanding element,
c  compute coefficient alpha
C
      IF(.NOT.YES$) THEN
         DO J4TEMP$=2,Tasks
            WORKpap(1)=WORKpap(1)+WORKpap(J4TEMP$)
         END DO
         CALL InitializeParallelEvent(EventB,Tasks)
         Outstanding=1
         Alpha=-WORKold(1)/WORKpap(1)
         CALL CompleteParallelEvent(EventA)
      END IF
C
C  [critical section] --- functions 5 and 6
C  distribute subsets
C
70    CALL LockonParallelLock(LockID)
      IF(Outstanding.GT.N) THEN
         CALL UnlockParallelLock(LockID)
         GO TO 80
      END IF
      J4TEMP$=Outstanding
      Outstanding=Outstanding+Subset56
      CALL UnlockParallelLock(LockID)
C
C  [parallellism] --- functions 5 and 6
C  compute a partial solution, and residual. 
C  size=Subset56*(3*N+2)
C
      DO K4TEMP$=J4TEMP$,MIN0(J4TEMP$+Subset56-1,N)
         X(K4TEMP$)=X(K4TEMP$)-ALPHA*P(K4TEMP$)
         DO M4TEMP$=1,N
            R(K4TEMP$)=R(K4TEMP$)+ALPHA*A(K4TEMP$,M4TEMP$)*P(M4TEMP$)
         END DO
      END DO
      GO TO 70
C
C  [task synchronization] --- functions 5 and 6
C  complete a cycle of event B
C
80    CALL PostParallelEvent(EventB)
      CALL IfCompleteParallelEvent(EventB,Yes$)
C
C  [sequential fragment] --- functions 5 and 6
C  initialize event A and outstanding element
C
      IF(.NOT.YES$) THEN
         CALL InitializeParallelEvent(EventA,Tasks)
         Outstanding=1
         CALL CompleteParallelEvent(EventB)
      END IF
C
C  [critical section] --- functions 7 and 8
C  distribute subset
C
      WORKnew(TaskIndex)=0.0
90    CALL LockonParallelLock(LockID)
      IF(Outstanding.GT.N) THEN
         CALL UnlockParallelLock(LockID)
         GO TO 100
      END IF
      J4TEMP$=Outstanding
      Outstanding=Outstanding+Subset78
      CALL UnlockParallelLock(LockID)
C
C  [parallelism] --- functions 7 and 8
C  compute a partial sum of norm, size=2*subset78
C
      DO K4TEMP$=J4TEMP$,MIN0(J4TEMP$+Subset78-1,N)
         WORKnew(TaskIndex)=WORKnew(TaskIndex)+R(K4TEMP$)*R(K4TEMP$)
      END DO
      GO TO 90
C
C  [task synchronization]
C  complete a cycle of event A
C
100   CALL PostParallelEvent(EventA)
      CALL IfCompleteParallelEvent(EventA,Yes$)
C
C  [sequential fragment]
C  sum residual,
C  and compute coefficient beta and initialize event B if necessary
C
      IF(.NOT.Yes$) THEN
         DO J4TEMP$=2,Tasks
            WORKnew(1)=WORKnew(1)+WORKnew(J4TEMP$)
         END DO
         IF(WORKnew(1).LT.Tolerance) THEN
            Done=.True.
         ELSE
            BETA=WORKnew(1)/WORKold(1)
            WORKold(1)=WORKnew(1)
         END IF
         Outstanding=1
         CALL InitializeParallelEvent(EventB,Tasks)
         CALL CompleteParallelEvent(EventA)
      END IF
C
C  exit the procedure
C
      IF(Done) RETURN
C   
C  [critical section] --- function 9
C  distribute subsets
C
110   CALL LockonParallelLock(LockID)
      IF(Outstanding.GT.N) THEN
         CALL UnlockParallelLock(LockID)
         GO TO 120
      END IF
      J4TEMP$=Outstanding
      Outstanding=Outstanding+Subset9
      CALL UnlockParallelLock(LockID)
C
C  [parallelism] --- function 9
C  compute directional vector, size=2*subset9
C
      DO K4TEMP$=J4TEMP$,MIN0(J4TEMP$+Subset9-1,N)
         P(K4TEMP$)=R(K4TEMP$)+BETA*P(K4TEMP$)
      END DO
      GO TO 110
C
C  [task synchronization] --- function 9
C  complete a cycle of event B
C
120   CALL PostParallelEvent(EventB)
      CALL IfCompleteParallelEvent(EventB,Yes$)
C
C  [sequential fragment] --- function 10
C  initialize event A and outstanding element
C
      IF(.NOT.YES$) THEN
         CALL InitializeParallelEvent(EventA,Tasks)         
         Outstanding=1
         CALL CompleteParallelEvent(EventB)
      END IF
C
C  advance to the next iteration
C
      WORKpap(TaskIndex)=0.0
      GO TO 50
C
      END

      REAL*4 FUNCTION RAN(SEED)
C
C
C  function to generate a pseudo-random number
C  (A)FORTRAN ENTRY: RAN(SEED)
C     1.RAN: <R4> return a random number
C     2.SEED: <I4> a seed (updated)
C
      INTEGER*4 SEED,A
      REAL*8 RTEMPA,RTEMPB
      DATA M/2147483647/,A/16807/
C
C  generate a random number
C
      RTEMPA=DFLOAT(SEED)*A
      ITEMP=INT(RTEMPA/DFLOAT(M))
      RTEMPB=DFLOAT(M)*DFLOAT(ITEMP)
      SEED=INT(RTEMPA-RTEMPB)
      RAN=DFLOAT(SEED)/DFLOAT(M)
C
      RETURN
      END
 
